#include "rot1_evaluator.h"
#include "util_sort.h"
#include "reg_binary_tree.h"
#include "reg_common.h"
#include "geometry.h"

using namespace reg::search;


ArcEvaluator::ArcEvaluator(
                           const reg::Matrix3X &X, const reg::Matrix3X &Y,
                           const reg::Vector &TH ):
    M_in(X), B_in(Y), TH_in(TH), _size(0)
{}


ArcEvaluator::~ArcEvaluator()
{
    if (_size>0)
    {
        for (int i=0; i<_size; i++)
        {
            delete tree_array[i];
        }
        delete []tree_array;
    }
}

int ArcEvaluator::sweep()
{
    //-----------------------------
    //        Sort
    //-----------------------------
    
    const int msize = (int)M_in.cols();
    const int bsize = (int)B_in.cols();

    IntervalTree **tree_array_tmp = new IntervalTree*[msize]();

    // Norm of points on XY-plane.
    Vector M_len(msize);
    int i;
    for(i=0; i<msize; i++)
    {
        tree_array_tmp[i] = NULL;
        double &x = M_in.x[3*i];
        double &y = M_in.x[3*i+1];
        
        M_len(i)=sqrt(x*x + y*y);
    }

    Vector B_len(bsize);
    for(i=0; i<bsize; i++)
    {
        double &x = B_in.x[3*i];
        double &y = B_in.x[3*i+1];
        
        B_len(i)=sqrt(x*x + y*y);
    }

    // Z-coordinate of points.

    Vector M_z(msize);
    Vector B_z(bsize);
    
    for(i=0; i<msize; i++)
    {
        M_z(i)=M_in(i,2);
    }

    for(i=0; i<bsize; i++)
    {
        B_z(i)=B_in(i,2);
    }
    
    // Azimuth of points.
    Vector M_azi(msize);
    Vector B_azi(bsize);

    for(i=0; i<msize; i++)
    {
        M_azi(i) = atan2(M_in(1,i), M_in(0,i));
    }

    for(i=0; i<bsize; i++)
    {
        B_azi(i) = atan2(B_in(1,i), B_in(0,i));
    }

    unsigned int *midx = reg::util::sort_index(M_len.x, msize);
    double *stmp = (double *)mxMalloc(msize*sizeof(double));

    //map
    Vector TH = TH_in;

    reg::util::sorted_by_index2(M_z.x  , midx, msize, stmp);
    reg::util::sorted_by_index2(M_azi.x, midx, msize, stmp);
    reg::util::sorted_by_index2(TH.x, midx, msize, stmp);

    mxFree(stmp);
    mxFree(midx);


    //-----------------------------
    //        Sweep
    //-----------------------------

    int lw, up;
    double maxth,dz,d,rth,dev,beg,end;
    double blen;
    //bool *hit;

    binarytree::payload newdata;
    binarytree::treeNode *status;

    maxth=0;
    status = NULL;
    for (i=0; i<msize; i++)
    {
        // Find maximum threshold
        if (TH(i)>maxth)
        {
            maxth = TH(i);
        }

        // Create plane sweep status.
        newdata.val = M_len(i);
        newdata.order = i;
        status = Insert(status, newdata, NULL);
    }

    // Start sweep.
    //hit = new bool[msize](); //(bool *)calloc(msize,sizeof(bool));

    int j;
    for (j=0; j<bsize; j++)
    {
        blen = B_len(j);
        lw = queryLower(status,blen-maxth,NULL);
        up = queryUpper(status,blen+maxth,NULL);

        for (i=lw;i<=up;i++)
        {
            dz = B_z(j)-M_z(i);
            if ( dz<-TH(i) || dz>TH(i)) /*|dz|>th[i]*/
            {
                continue;
            }
            d = blen-M_len(i); /*d = fabs(blen-MLEN[i]);*/
            rth = sqrt(TH(i)*TH(i)-dz*dz);
            if ( d<-rth || d>rth)  /* |d|>rth  */
            {
                continue;
            }

            if (tree_array_tmp[i]==NULL)
            {
                tree_array_tmp[i] = new IntervalTree();
            }

            if (M_len(i)<=DUMMY_PRECISION)
            {
                tree_array_tmp[i]->insert(0, TWOPI);
            }
            else
            {
                dev = reg::geometry::circleintersection(M_len(i), blen, rth);

                if (fabs(dev-PI) <= DUMMY_PRECISION )
                {
                    /*That could be improved by instead of insert adding 1 to the final quality*/
                    tree_array_tmp[i]->insert(0, TWOPI);
                }
                else
                {
                    beg = fmod(B_azi(j)-dev-M_azi(i), TWOPI);
                    if (beg<0)
                    {
                        beg += TWOPI;
                    }
                    end = fmod(B_azi(j)+dev-M_azi(i), TWOPI);
                    if (end<0)
                    {
                        end += TWOPI;
                    }
                    if (end>=beg)
                    {
                        tree_array_tmp[i]->insert(beg, end);
                    }
                    else
                    {
                        tree_array_tmp[i]->insert(beg, TWOPI);
                        tree_array_tmp[i]->insert(0, end);
                    }
                }
            }
            //hit[i]=true;
        }
    }

    free_Binarytree(status);

    _size = 0;


    for (i=0; i<msize; i++)
    {
        if (tree_array_tmp[i]!=NULL) //(hit[i])
        {
            _size++;
        }
    }

    if(_size>0)
    {
        tree_array = new IntervalTree*[_size];
        j=0;
        for (i=0; i<msize; i++)
        {
            if (tree_array_tmp[i]!=NULL)
            {
                tree_array[j++] = tree_array_tmp[i];
            }
        }
    }

    delete []tree_array_tmp;

    return _size;
}


int ArcEvaluator::size() const
{
    return _size;
}

int ArcEvaluator::evalUpperBound(SSR ssr, int lwbnd) const
{
    const double lw = ssr.lw;
    const double up = ssr.up;
    int bnd=0;

    //Second condition:  bnd + (max remaining) > lwbnd
    for (int i=0; i<_size && (bnd + _size-i > lwbnd); i++)
    {
        bnd += tree_array[i]->matchInter(lw, up);
    }

    return bnd;
}

int ArcEvaluator::evalUpperBound(SSR ssr, int lwbnd,
                           int *matchList, int matchListSize,
                           std::vector<bool> &matches) const
{
    mxAssert(matchListSize>=0 && matchListSize<=_size, "");

    const double lw = ssr.lw;
    const double up = ssr.up;
    int bnd=0;

    for (int i=0; i<matchListSize && (bnd+matchListSize-i > lwbnd); i++)
    {
        if(tree_array[matchList[i]]->matchInter(lw, up))
        {
            bnd++;
            matches[i]=1;
        }
        else
        {
            matches[i]=0;
        }
    }

    return bnd;
}


int ArcEvaluator::evalLowerBound(SSR ssr) const
{
    const double ang = .5*(ssr.lw + ssr.up);
    int qual=0;

    for (int i=0; i<_size; i++)
    {
        qual += tree_array[i]->matchValue(ang);
    }
    return qual;
}

int ArcEvaluator::evalLowerBound(
        SSR ssr,
        int *matchList, int matchListSize) const
{
    mxAssert(matchListSize>=0 && matchListSize<=_size, "");

    const double ang = .5*(ssr.lw + ssr.up);
    int qual=0;

    for (int i=0; i<matchListSize; i++)
    {
        qual += tree_array[matchList[i]]->matchValue(ang);
    }

    return qual;
}

int ArcEvaluator::evalLowerBound(SSR ssr, int *matchList,
                                         int matchListSize,
                                         std::vector<bool> &matches) const
{
    mxAssert(matchListSize>=0 && matchListSize<=_size, "");

    const double ang = .5*(ssr.lw + ssr.up);
    int qual=0;

    for (int i=0; i<matchListSize; i++)
    {
        if(tree_array[matchList[i]]->matchValue(ang))
        {
            matches[i]=1;
            qual++;
        }
        else
        {
            matches[i]=0;
        }
    }

    return qual;
}

